/*
 * kgraft_patch_bsc955837
 *
 * Fix for CVE-2013-7446, bsc#955837
 * Fix for bsc#973570
 *
 *  Upstream commits:
 *  7d267278a9ece963d77eefec61630223fce08c6c
 *  a5527dda344fff0514b7989ef7a755729769daa1
 *
 *  SLE12 commits:
 *  6f55ce76866e80c205183e08136f7b1b4d263cef (3.12.52 stable)
 *  9cc7d785fa9868fa71a86cf5a79eb67a28eea039 (3.12.56 -stable)
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/net.h>
#include <linux/poll.h>
#include <net/af_unix.h>
#include <net/tcp_states.h>

static unsigned int *kgr_unix_tot_inflight;
static void (*kgr_unix_gc)(void);
static atomic_long_t *kgr_unix_nr_socks;
static struct proto *kgr_unix_proto;
static void (*kgr_unix_write_space)(struct sock *sk);
static void (*kgr__unix_insert_socket)(struct hlist_head *list, struct sock *sk);
static int (*kgr_unix_mkname)(struct sockaddr_un *sunaddr, int len, unsigned int *hashp);
static struct sock *(*kgr__unix_find_socket_byname)(struct net *net, struct sockaddr_un *sunname, int len, int type, unsigned int hash);
static struct sock *(*kgr_unix_find_other)(struct net *net, struct sockaddr_un *sunname, int len, int type, unsigned int hash, int *error);
static void (*kgr_unix_dgram_disconnected)(struct sock *sk, struct sock *other);
static int (*kgr_unix_scm_to_skb)(struct scm_cookie *scm, struct sk_buff *skb, bool send_fds);
static long (*kgr_unix_wait_for_peer)(struct sock *other, long timeo);
static void (*kgr_maybe_add_creds)(struct sk_buff *skb, const struct socket *sock, const struct sock *other);
static void (*kgr_wait_for_unix_gc)(void);
static int (*kgr_unix_bind)(struct socket *sock, struct sockaddr *uaddr, int addr_len);
static int (*kgr_unix_socketpair)(struct socket *socka, struct socket *sockb);
static int (*kgr_unix_getname)(struct socket *sock, struct sockaddr *uaddr, int *uaddr_len, int peer);
static int (*kgr_unix_ioctl)(struct socket *sock, unsigned int cmd, unsigned long arg);
static int (*kgr_unix_shutdown)(struct socket *sock, int mode);
static int (*kgr_unix_dgram_recvmsg)(struct kiocb *iocb, struct socket *sock, struct msghdr *msg, size_t size, int flags);
static int (*kgr_unix_set_peek_off)(struct sock *sk, int val);
static const struct proto_ops *kgr_unix_stream_ops;
static const struct proto_ops *kgr_unix_seqpacket_ops;

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "unix_tot_inflight", (void *)&kgr_unix_tot_inflight },
	{ "unix_gc", (void *)&kgr_unix_gc },
	{ "unix_nr_socks", (void *)&kgr_unix_nr_socks },
	{ "unix_proto", (void *)&kgr_unix_proto },
	{ "unix_write_space", (void *)&kgr_unix_write_space },
	{ "__unix_insert_socket", (void *)&kgr__unix_insert_socket },
	{ "unix_mkname", (void *)&kgr_unix_mkname },
	{ "__unix_find_socket_byname", (void *)&kgr__unix_find_socket_byname },
	{ "unix_find_other", (void *)&kgr_unix_find_other },
	{ "unix_dgram_disconnected", (void *)&kgr_unix_dgram_disconnected },
	{ "unix_scm_to_skb", (void *)&kgr_unix_scm_to_skb },
	{ "unix_wait_for_peer", (void *)&kgr_unix_wait_for_peer },
	{ "maybe_add_creds", (void *)&kgr_maybe_add_creds },
	{ "wait_for_unix_gc", (void *)&kgr_wait_for_unix_gc },
	{ "unix_bind", (void *)&kgr_unix_bind },
	{ "unix_socketpair", (void *)&kgr_unix_socketpair },
	{ "unix_getname", (void *)&kgr_unix_getname },
	{ "unix_ioctl", (void *)&kgr_unix_ioctl },
	{ "unix_shutdown", (void *)&kgr_unix_shutdown },
	{ "unix_dgram_recvmsg", (void *)&kgr_unix_dgram_recvmsg },
	{ "unix_set_peek_off", (void *)&kgr_unix_set_peek_off },
	{ "unix_stream_ops", (void *)&kgr_unix_stream_ops },
	{ "unix_seqpacket_ops", (void *)&kgr_unix_seqpacket_ops },
};

/* New ops to ensure consistency for new sockets */
static struct proto_ops kgr_unix_dgram_ops = {
	.family =	PF_UNIX,
	.owner =	THIS_MODULE,
	.accept =	sock_no_accept,
	.listen =	sock_no_listen,
	.setsockopt =	sock_no_setsockopt,
	.getsockopt =	sock_no_getsockopt,
	.mmap =		sock_no_mmap,
	.sendpage =	sock_no_sendpage,
};

#define unix_peer(sk) (unix_sk(sk)->peer)

/* New KABI structure */
struct unix_peer_wake {
	wait_queue_t		wq;
	struct unix_sock	*sk;
};

/* New KABI function */
static inline wait_queue_t *kgr_get_peer_wake(struct unix_sock *usk)
{
	struct unix_peer_wake *pw;

	pw = *((struct unix_peer_wake **)&usk->sk.__sk_common.skc_daddr);
	return &pw->wq;
}

/* New KABI function */
static inline void kgr_set_peer_wake(struct unix_sock *usk,
				     struct unix_peer_wake *pw)
{
	pw->sk = usk;
	*((struct unix_peer_wake **)&usk->sk.__sk_common.skc_daddr) = pw;
}

/* New KABI function */
static inline struct unix_sock *kgr_get_peer_wake_sk(wait_queue_t *q)
{
	struct unix_peer_wake *pw = container_of(q, struct unix_peer_wake, wq);

	return pw->sk;
}

/* New KABI function */
static inline void kgr_unix_release_peer_wake(struct unix_sock *usk)
{
	kfree(*((struct unix_peer_wake **)&usk->sk.__sk_common.skc_daddr));
}

static void kgr__unix_remove_socket(struct sock *sk)
{
	sk_del_node_init(sk);
}

static inline void kgr_unix_remove_socket(struct sock *sk)
{
	spin_lock(&unix_table_lock);
	kgr__unix_remove_socket(sk);
	spin_unlock(&unix_table_lock);
}

static inline void kgr_unix_release_addr(struct unix_address *addr)
{
	if (atomic_dec_and_test(&addr->refcnt))
		kfree(addr);
}

static inline void kgr_unix_insert_socket(struct hlist_head *list, struct sock *sk)
{
	spin_lock(&unix_table_lock);
	kgr__unix_insert_socket(list, sk);
	spin_unlock(&unix_table_lock);
}

static struct hlist_head *kgr_unix_sockets_unbound(void *addr)
{
	unsigned long hash = (unsigned long)addr;

	hash ^= hash >> 16;
	hash ^= hash >> 8;
	hash %= UNIX_HASH_SIZE;
	return &unix_socket_table[UNIX_HASH_SIZE + hash];
}

static inline unsigned int kgr_unix_hash_fold(__wsum n)
{
	unsigned int hash = (__force unsigned int)csum_fold(n);

	hash ^= hash>>8;
	return hash&(UNIX_HASH_SIZE-1);
}

static inline int kgr_unix_our_peer(struct sock *sk, struct sock *osk)
{
	return unix_peer(osk) == sk;
}

static inline int kgr_unix_may_send(struct sock *sk, struct sock *osk)
{
	return unix_peer(osk) == NULL || kgr_unix_our_peer(sk, osk);
}

#ifdef CONFIG_SECURITY_NETWORK
static void kgr_unix_get_secdata(struct scm_cookie *scm, struct sk_buff *skb)
{
	memcpy(UNIXSID(skb), &scm->secid, sizeof(u32));
}
#else
static inline void kgr_unix_get_secdata(struct scm_cookie *scm, struct sk_buff *skb)
{ }
#endif /* CONFIG_SECURITY_NETWORK */

static inline int kgr_unix_recvq_full(struct sock const *sk)
{
	return skb_queue_len(&sk->sk_receive_queue) > sk->sk_max_ack_backlog;
}

/* New function */
static int kgr_unix_dgram_peer_wake_relay(wait_queue_t *q, unsigned mode,
	int flags, void *key)
{
	struct unix_sock *u;
	wait_queue_head_t *u_sleep;

	/*
	 * KGR - KABI fix
	 * -u = container_of(q, struct unix_sock, peer_wake);
	 * +next row
	 */
	u = kgr_get_peer_wake_sk(q);

	/*
	 * KGR - KABI fix
	 * -__remove_wait_queue(&unix_sk(u->peer_wake.private)->peer_wait,
	 * +next row
	 */
	__remove_wait_queue(&unix_sk(kgr_get_peer_wake(u)->private)->peer_wait,
			    q);
	/*
	 * KGR - KABI fix
	 * -u->peer_wake.private = NULL;
	 * +next row
	 */
	kgr_get_peer_wake(u)->private = NULL;

	/* relaying can only happen while the wq still exists */
	u_sleep = sk_sleep(&u->sk);
	if (u_sleep)
		wake_up_interruptible_poll(u_sleep, key);

	return 0;
}

/* New function */
static int kgr_unix_dgram_peer_wake_connect(struct sock *sk, struct sock *other)
{
	struct unix_sock *u, *u_other;
	int rc;

	u = unix_sk(sk);
	u_other = unix_sk(other);
	rc = 0;
	spin_lock(&u_other->peer_wait.lock);

	/*
	 * KGR - KABI fix
	 *  -if (!u->peer_wake.private) {
	 *  -	u->peer_wake.private = other;
	 *  -	__add_wait_queue(&u_other->peer_wait, &u->peer_wake);
	 *  +next 3 rows
	 */
	if (!kgr_get_peer_wake(u)->private) {
		kgr_get_peer_wake(u)->private = other;
		__add_wait_queue(&u_other->peer_wait, kgr_get_peer_wake(u));

		rc = 1;
	}

	spin_unlock(&u_other->peer_wait.lock);
	return rc;
}

/* New function */
static void kgr_unix_dgram_peer_wake_disconnect(struct sock *sk,
	struct sock *other)
{
	struct unix_sock *u, *u_other;

	u = unix_sk(sk);
	u_other = unix_sk(other);
	spin_lock(&u_other->peer_wait.lock);

	/*
	 * KGR - KABI fix
	 *
	 * -if (u->peer_wake.private == other) {
	 * -	__remove_wait_queue(&u_other->peer_wait, &u->peer_wake);
	 * -	u->peer_wake.private = NULL;
	 * -}
	 */
	if (kgr_get_peer_wake(u)->private == other) {
		__remove_wait_queue(&u_other->peer_wait, kgr_get_peer_wake(u));
		kgr_get_peer_wake(u)->private = NULL;
	}

	spin_unlock(&u_other->peer_wait.lock);
}

/* New function */
static void kgr_unix_dgram_peer_wake_disconnect_wakeup(struct sock *sk,
	struct sock *other)
{
	kgr_unix_dgram_peer_wake_disconnect(sk, other);
	wake_up_interruptible_poll(sk_sleep(sk),
				   POLLOUT |
				   POLLWRNORM |
				   POLLWRBAND);
}

/* New function */
static int kgr_unix_dgram_peer_wake_me(struct sock *sk, struct sock *other)
{
	int connected;

	connected = kgr_unix_dgram_peer_wake_connect(sk, other);

	if (kgr_unix_recvq_full(other))
		return 1;

	if (connected)
		kgr_unix_dgram_peer_wake_disconnect(sk, other);

	return 0;
}

/*
 * Patched function
 * Not kgraft patched to ensure consistency of struct sock.
 */
static void kgr_unix_release_sock(struct sock *sk, int embrion)
{
	struct unix_sock *u = unix_sk(sk);
	struct path path;
	struct sock *skpair;
	struct sk_buff *skb;
	int state;

	kgr_unix_remove_socket(sk);

	/* Clear state */
	unix_state_lock(sk);
	sock_orphan(sk);
	sk->sk_shutdown = SHUTDOWN_MASK;
	path	     = u->path;
	u->path.dentry = NULL;
	u->path.mnt = NULL;
	state = sk->sk_state;
	sk->sk_state = TCP_CLOSE;
	unix_state_unlock(sk);

	wake_up_interruptible_all(&u->peer_wait);

	skpair = unix_peer(sk);

	if (skpair != NULL) {
		if (sk->sk_type == SOCK_STREAM || sk->sk_type == SOCK_SEQPACKET) {
			unix_state_lock(skpair);
			/* No more writes */
			skpair->sk_shutdown = SHUTDOWN_MASK;
			if (!skb_queue_empty(&sk->sk_receive_queue) || embrion)
				skpair->sk_err = ECONNRESET;
			unix_state_unlock(skpair);
			skpair->sk_state_change(skpair);
			sk_wake_async(skpair, SOCK_WAKE_WAITD, POLL_HUP);
		}

		/* Fix CVE-2013-7446 */
		kgr_unix_dgram_peer_wake_disconnect(sk, skpair);
		sock_put(skpair); /* It may now die */
		unix_peer(sk) = NULL;
	}

	/* Try to flush out this socket. Throw out buffers at least */

	while ((skb = skb_dequeue(&sk->sk_receive_queue)) != NULL) {
		if (state == TCP_LISTEN)
			kgr_unix_release_sock(skb->sk, 1);
		/* passed fds are erased in the kfree_skb hook	      */
		kfree_skb(skb);
	}

	if (path.dentry)
		path_put(&path);

	sock_put(sk);

	/* ---- Socket is dead now and most probably destroyed ---- */

	/*
	 * Fixme: BSD difference: In BSD all sockets connected to us get
	 *	  ECONNRESET and we die on the spot. In Linux we behave
	 *	  like files and pipes do and wait for the last
	 *	  dereference.
	 *
	 * Can't we simply set sock->err?
	 *
	 *	  What the above comment does talk about? --ANK(980817)
	 */

	if (*kgr_unix_tot_inflight)
		kgr_unix_gc();		/* Garbage collect fds */
}

/*
 * Patched function
 * Not kgraft patched to ensure consistency of struct sock.
 */
static int kgr_unix_release(struct socket *sock)
{
	struct sock *sk = sock->sk;

	if (!sk)
		return 0;

	kgr_unix_release_sock(sk, 0);
	sock->sk = NULL;

	return 0;
}

/*
 * Changed function - due to KABI
 * Not kgraft patched to ensure consistency of struct sock.
 */
static void kgr_unix_sock_destructor(struct sock *sk)
{
	struct unix_sock *u = unix_sk(sk);

	skb_queue_purge(&sk->sk_receive_queue);

	WARN_ON(atomic_read(&sk->sk_wmem_alloc));
	WARN_ON(!sk_unhashed(sk));
	WARN_ON(sk->sk_socket);
	if (!sock_flag(sk, SOCK_DEAD)) {
		printk(KERN_INFO "Attempt to release alive unix socket: %p\n", sk);
		return;
	}

	if (u->addr)
		kgr_unix_release_addr(u->addr);
	/* KGR - KABI fix, +next row */
	kgr_unix_release_peer_wake(u);

	atomic_long_dec(kgr_unix_nr_socks);
	local_bh_disable();
	sock_prot_inuse_add(sock_net(sk), sk->sk_prot, -1);
	local_bh_enable();
#ifdef UNIX_REFCNT_DEBUG
	printk(KERN_DEBUG "UNIX %p is destroyed, %ld are still alive.\n", sk,
		atomic_long_read(kgr_unix_nr_socks));
#endif
}

/*
 * Patched function
 * Not kgraft patched to ensure consistency of struct sock.
 */
static struct sock *kgr_unix_create1(struct net *net, struct socket *sock)
{
	struct sock *sk = NULL;
	struct unix_sock *u;
	/* KGR - KABI fix, +next row and kmalloc */
	struct unix_peer_wake *pw;

	pw = kmalloc(sizeof(*pw), GFP_KERNEL);
	if (!pw)
		return NULL;

	atomic_long_inc(kgr_unix_nr_socks);
	if (atomic_long_read(kgr_unix_nr_socks) > 2 * get_max_files())
		goto out;

	sk = sk_alloc(net, PF_UNIX, GFP_KERNEL, kgr_unix_proto);
	if (!sk)
		goto out;

	sock_init_data(sock, sk);
	/*
	 * KGR - CONFIG_LOCKDEP is off, so lockdep_set_class is basically a
	 * nop. Remove it.
	 * -lockdep_set_class(&sk->sk_receive_queue.lock,
	 * -			&af_unix_sk_receive_queue_lock_key);
	 */

	sk->sk_write_space	= kgr_unix_write_space;
	sk->sk_max_ack_backlog	= net->unx.sysctl_max_dgram_qlen;
	sk->sk_destruct		= kgr_unix_sock_destructor;
	u	  = unix_sk(sk);
	u->path.dentry = NULL;
	u->path.mnt = NULL;
	spin_lock_init(&u->lock);
	atomic_long_set(&u->inflight, 0);
	INIT_LIST_HEAD(&u->link);
	mutex_init(&u->readlock); /* single task reading lock */
	init_waitqueue_head(&u->peer_wait);
	/*
	 * Fix CVE-2013-7446
	 * +init_waitqueue_func_entry(&u->peer_wake, unix_dgram_peer_wake_relay);
	 * Removed by KABI fix
	 * Added next two rows
	 */
	kgr_set_peer_wake(u, pw);
	init_waitqueue_func_entry(kgr_get_peer_wake(u), kgr_unix_dgram_peer_wake_relay);
	kgr_unix_insert_socket(kgr_unix_sockets_unbound(sk), sk);
out:
	if (sk == NULL) {
		atomic_long_dec(kgr_unix_nr_socks);
		/* KGR - KABI fix, +next row */
		kfree(pw);
	} else {
		local_bh_disable();
		sock_prot_inuse_add(sock_net(sk), sk->sk_prot, 1);
		local_bh_enable();
	}
	return sk;
}

/* Patched function */
int kgr_unix_create(struct net *net, struct socket *sock, int protocol, int kern)
{
	if (protocol && protocol != PF_UNIX)
		return -EPROTONOSUPPORT;

	sock->state = SS_UNCONNECTED;

	switch (sock->type) {
	case SOCK_STREAM:
		sock->ops = kgr_unix_stream_ops;
		break;
		/*
		 *	Believe it or not BSD has AF_UNIX, SOCK_RAW though
		 *	nothing uses it.
		 */
	case SOCK_RAW:
		sock->type = SOCK_DGRAM;
	case SOCK_DGRAM:
		sock->ops = &kgr_unix_dgram_ops;
		break;
	case SOCK_SEQPACKET:
		/* KGR - use new ops for a new socket */
		sock->ops = kgr_unix_seqpacket_ops;
		break;
	default:
		return -ESOCKTNOSUPPORT;
	}

	return kgr_unix_create1(net, sock) ? 0 : -ENOMEM;
}

/* isra optimized function */
static int kgr_unix_autobind(struct socket *sock)
{
	struct sock *sk = sock->sk;
	struct net *net = sock_net(sk);
	struct unix_sock *u = unix_sk(sk);
	static u32 ordernum = 1;
	struct unix_address *addr;
	int err;
	unsigned int retries = 0;

	err = mutex_lock_interruptible(&u->readlock);
	if (err)
		return err;

	err = 0;
	if (u->addr)
		goto out;

	err = -ENOMEM;
	addr = kzalloc(sizeof(*addr) + sizeof(short) + 16, GFP_KERNEL);
	if (!addr)
		goto out;

	addr->name->sun_family = AF_UNIX;
	atomic_set(&addr->refcnt, 1);

retry:
	addr->len = sprintf(addr->name->sun_path+1, "%05x", ordernum) + 1 + sizeof(short);
	addr->hash = kgr_unix_hash_fold(csum_partial(addr->name, addr->len, 0));

	spin_lock(&unix_table_lock);
	ordernum = (ordernum+1)&0xFFFFF;

	if (kgr__unix_find_socket_byname(net, addr->name, addr->len, sock->type,
				      addr->hash)) {
		spin_unlock(&unix_table_lock);
		/*
		 * __unix_find_socket_byname() may take long time if many names
		 * are already in use.
		 */
		cond_resched();
		/* Give up if all names seems to be in use. */
		if (retries++ == 0xFFFFF) {
			err = -ENOSPC;
			kfree(addr);
			goto out;
		}
		goto retry;
	}
	addr->hash ^= sk->sk_type;

	kgr__unix_remove_socket(sk);
	u->addr = addr;
	kgr__unix_insert_socket(&unix_socket_table[addr->hash], sk);
	spin_unlock(&unix_table_lock);
	err = 0;

out:	mutex_unlock(&u->readlock);
	return err;
}

/* Inlined function */
static void kgr_unix_state_double_lock(struct sock *sk1, struct sock *sk2)
{
	if (unlikely(sk1 == sk2) || !sk2) {
		unix_state_lock(sk1);
		return;
	}
	if (sk1 < sk2) {
		unix_state_lock(sk1);
		unix_state_lock_nested(sk2);
	} else {
		unix_state_lock(sk2);
		unix_state_lock_nested(sk1);
	}
}

/* Inlined function */
static void kgr_unix_state_double_unlock(struct sock *sk1, struct sock *sk2)
{
	if (unlikely(sk1 == sk2) || !sk2) {
		unix_state_unlock(sk1);
		return;
	}
	unix_state_unlock(sk1);
	unix_state_unlock(sk2);
}

/*
 * Patched function
 * Not kgraft patched to ensure consistency of struct sock.
 */
static int kgr_unix_dgram_connect(struct socket *sock, struct sockaddr *addr,
	int alen, int flags)
{
	struct sock *sk = sock->sk;
	struct net *net = sock_net(sk);
	struct sockaddr_un *sunaddr = (struct sockaddr_un *)addr;
	struct sock *other;
	unsigned int hash;
	int err;

	if (addr->sa_family != AF_UNSPEC) {
		err = kgr_unix_mkname(sunaddr, alen, &hash);
		if (err < 0)
			goto out;
		alen = err;

		if (test_bit(SOCK_PASSCRED, &sock->flags) &&
		    !unix_sk(sk)->addr && (err = kgr_unix_autobind(sock)) != 0)
			goto out;

restart:
		other = kgr_unix_find_other(net, sunaddr, alen, sock->type, hash, &err);
		if (!other)
			goto out;

		kgr_unix_state_double_lock(sk, other);

		/* Apparently VFS overslept socket death. Retry. */
		if (sock_flag(other, SOCK_DEAD)) {
			kgr_unix_state_double_unlock(sk, other);
			sock_put(other);
			goto restart;
		}

		err = -EPERM;
		if (!kgr_unix_may_send(sk, other))
			goto out_unlock;

		err = security_unix_may_send(sk->sk_socket, other->sk_socket);
		if (err)
			goto out_unlock;

	} else {
		/*
		 *	1003.1g breaking connected state with AF_UNSPEC
		 */
		other = NULL;
		kgr_unix_state_double_lock(sk, other);
	}

	/*
	 * If it was connected, reconnect.
	 */
	if (unix_peer(sk)) {
		struct sock *old_peer = unix_peer(sk);
		unix_peer(sk) = other;
		/* Fix CVE-2013-7446, +next row */
		kgr_unix_dgram_peer_wake_disconnect_wakeup(sk, old_peer);

		kgr_unix_state_double_unlock(sk, other);

		if (other != old_peer)
			kgr_unix_dgram_disconnected(sk, old_peer);
		sock_put(old_peer);
	} else {
		unix_peer(sk) = other;
		kgr_unix_state_double_unlock(sk, other);
	}
	return 0;

out_unlock:
	kgr_unix_state_double_unlock(sk, other);
	sock_put(other);
out:
	return err;
}

/*
 * Patched function
 * Not kgraft patched to ensure consistency of struct sock.
 */
static int kgr_unix_dgram_sendmsg(struct kiocb *kiocb, struct socket *sock,
	struct msghdr *msg, size_t len)
{
	struct sock_iocb *siocb = kiocb_to_siocb(kiocb);
	struct sock *sk = sock->sk;
	struct net *net = sock_net(sk);
	struct unix_sock *u = unix_sk(sk);
	struct sockaddr_un *sunaddr = msg->msg_name;
	struct sock *other = NULL;
	int namelen = 0; /* fake GCC */
	int err;
	unsigned int hash;
	struct sk_buff *skb;
	long timeo;
	struct scm_cookie tmp_scm;
	int max_level;
	int data_len = 0;
	/* Fix CVE-2013-7446, +next row */
	int sk_locked;

	if (NULL == siocb->scm)
		siocb->scm = &tmp_scm;
	kgr_wait_for_unix_gc();
	err = scm_send(sock, msg, siocb->scm, false);
	if (err < 0)
		return err;

	err = -EOPNOTSUPP;
	if (msg->msg_flags&MSG_OOB)
		goto out;

	if (msg->msg_namelen) {
		err = kgr_unix_mkname(sunaddr, msg->msg_namelen, &hash);
		if (err < 0)
			goto out;
		namelen = err;
	} else {
		sunaddr = NULL;
		err = -ENOTCONN;
		other = unix_peer_get(sk);
		if (!other)
			goto out;
	}

	if (test_bit(SOCK_PASSCRED, &sock->flags) && !u->addr
	    && (err = kgr_unix_autobind(sock)) != 0)
		goto out;

	err = -EMSGSIZE;
	if (len > sk->sk_sndbuf - 32)
		goto out;

	if (len > SKB_MAX_ALLOC)
		data_len = min_t(size_t,
				 len - SKB_MAX_ALLOC,
				 MAX_SKB_FRAGS * PAGE_SIZE);

	skb = sock_alloc_send_pskb(sk, len - data_len, data_len,
				   msg->msg_flags & MSG_DONTWAIT, &err,
				   PAGE_ALLOC_COSTLY_ORDER);
	if (skb == NULL)
		goto out;

	err = kgr_unix_scm_to_skb(siocb->scm, skb, true);
	if (err < 0)
		goto out_free;
	max_level = err + 1;
	kgr_unix_get_secdata(siocb->scm, skb);

	skb_put(skb, len - data_len);
	skb->data_len = data_len;
	skb->len = len;
	err = skb_copy_datagram_from_iovec(skb, 0, msg->msg_iov, 0, len);
	if (err)
		goto out_free;

	timeo = sock_sndtimeo(sk, msg->msg_flags & MSG_DONTWAIT);

restart:
	if (!other) {
		err = -ECONNRESET;
		if (sunaddr == NULL)
			goto out_free;

		other = kgr_unix_find_other(net, sunaddr, namelen, sk->sk_type,
					hash, &err);
		if (other == NULL)
			goto out_free;
	}

	if (sk_filter(other, skb) < 0) {
		/* Toss the packet but do not return any error to the sender */
		err = len;
		goto out_free;
	}

	/* Fix CVE-2013-7446, +next row */
	sk_locked = 0;
	unix_state_lock(other);
	/* Fix CVE-2013-7446, +next row */
restart_locked:
	err = -EPERM;
	if (!kgr_unix_may_send(sk, other))
		goto out_unlock;

	/*
	 * Fix CVE-2013-7446
	 *  -if (sock_flag(other, SOCK_DEAD)) {
	 *  +next row
	 */
	if (unlikely(sock_flag(other, SOCK_DEAD))) {
		/*
		 *	Check with 1003.1g - what should
		 *	datagram error
		 */
		unix_state_unlock(other);
		sock_put(other);

		/* Fix CVE-2013-7446, +next two rows */
		if (!sk_locked)
			unix_state_lock(sk);

		err = 0;
		/*
		 * Fix CVE-2013-7446
		 *  -unix_state_lock(sk);
		 */
		if (unix_peer(sk) == other) {
			unix_peer(sk) = NULL;
			/* Fix CVE-2013-7446, +next row */
			kgr_unix_dgram_peer_wake_disconnect_wakeup(sk, other);

			unix_state_unlock(sk);

			kgr_unix_dgram_disconnected(sk, other);
			sock_put(other);
			err = -ECONNREFUSED;
		} else {
			unix_state_unlock(sk);
		}

		other = NULL;
		if (err)
			goto out_free;
		goto restart;
	}

	err = -EPIPE;
	if (other->sk_shutdown & RCV_SHUTDOWN)
		goto out_unlock;

	if (sk->sk_type != SOCK_SEQPACKET) {
		err = security_unix_may_send(sk->sk_socket, other->sk_socket);
		if (err)
			goto out_unlock;
	}

	/*
	 * Fix CVE-2013-7446
	 *  -if (unix_peer(other) != sk && kgr_unix_recvq_full(other)) {
	 *  -	if (!timeo) {
	 *  -		err = -EAGAIN;
	 *  -		goto out_unlock;
	 *  +next 9 rows
	 */
	/*
	 * Fix bsc#973570
	 *  -if (unlikely(unix_peer(other) != sk && kgr_unix_recvq_full(other))) {
	 *  +next 6 rows
	 */
	/* other == sk && unix_peer(other) != sk if
	 * - unix_peer(sk) == NULL, destination address bound to sk
	 * - unix_peer(sk) == sk by time of get but disconnected before lock
	 */
	if (other != sk &&
	    unlikely(unix_peer(other) != sk && kgr_unix_recvq_full(other))) {

		if (timeo) {
			timeo = kgr_unix_wait_for_peer(other, timeo);

			err = sock_intr_errno(timeo);
			if (signal_pending(current))
				goto out_free;

			goto restart;
		}

		/*
		 * Fix CVE-2013-7446
		 *  -timeo = kgr_unix_wait_for_peer(other, timeo);
		 *  +next 4 rows
		 */
		if (!sk_locked) {
			unix_state_unlock(other);
			kgr_unix_state_double_lock(sk, other);
		}

		/*
		 * Fix CVE-2013-7446
		 *  -err = sock_intr_errno(timeo);
		 *  -if (signal_pending(current))
		 *  -	goto out_free;
		 *  +next 6 rows
		 */
		if (unix_peer(sk) != other ||
		    kgr_unix_dgram_peer_wake_me(sk, other)) {
			err = -EAGAIN;
			sk_locked = 1;
			goto out_unlock;
		}

		/*
		 * Fix CVE-2013-7446
		 *  -goto restart;
		 *  +next 4 rows
		 */
		if (!sk_locked) {
			sk_locked = 1;
			goto restart_locked;
		}
	}

	/*
	 * Fix CVE-2013-7446
	 *  +next 3 rows
	 */
	if (unlikely(sk_locked))
		unix_state_unlock(sk);

	if (sock_flag(other, SOCK_RCVTSTAMP))
		__net_timestamp(skb);
	kgr_maybe_add_creds(skb, sock, other);
	skb_queue_tail(&other->sk_receive_queue, skb);
	if (max_level > unix_sk(other)->recursion_level)
		unix_sk(other)->recursion_level = max_level;
	unix_state_unlock(other);
	other->sk_data_ready(other, len);
	sock_put(other);
	scm_destroy(siocb->scm);
	return len;

out_unlock:
	/*
	 * Fix CVE-2013-7446
	 *  +next 2 rows
	 */
	if (sk_locked)
		unix_state_unlock(sk);
	unix_state_unlock(other);
out_free:
	kfree_skb(skb);
out:
	if (other)
		sock_put(other);
	scm_destroy(siocb->scm);
	return err;
}

static inline int kgr_unix_writable(struct sock *sk)
{
	return (atomic_read(&sk->sk_wmem_alloc) << 2) <= sk->sk_sndbuf;
}

/*
 * Patched function
 * Not kgraft patched to ensure consistency of struct sock.
 */
static unsigned int kgr_unix_dgram_poll(struct file *file, struct socket *sock,
	poll_table *wait)
{
	struct sock *sk = sock->sk, *other;
	unsigned int mask, writable;

	sock_poll_wait(file, sk_sleep(sk), wait);
	mask = 0;

	/* exceptional events? */
	if (sk->sk_err || !skb_queue_empty(&sk->sk_error_queue))
		mask |= POLLERR |
			(sock_flag(sk, SOCK_SELECT_ERR_QUEUE) ? POLLPRI : 0);

	if (sk->sk_shutdown & RCV_SHUTDOWN)
		mask |= POLLRDHUP | POLLIN | POLLRDNORM;
	if (sk->sk_shutdown == SHUTDOWN_MASK)
		mask |= POLLHUP;

	/* readable? */
	if (!skb_queue_empty(&sk->sk_receive_queue))
		mask |= POLLIN | POLLRDNORM;

	/* Connection-based need to check for termination and startup */
	if (sk->sk_type == SOCK_SEQPACKET) {
		if (sk->sk_state == TCP_CLOSE)
			mask |= POLLHUP;
		/* connection hasn't started yet? */
		if (sk->sk_state == TCP_SYN_SENT)
			return mask;
	}

	/* No write status requested, avoid expensive OUT tests. */
	if (!(poll_requested_events(wait) & (POLLWRBAND|POLLWRNORM|POLLOUT)))
		return mask;

	writable = kgr_unix_writable(sk);
	/*
	 * Fix CVE-2013-7446
	 *  -other = unix_peer_get(sk);
	 *  -if (other) {
	 *  -	if (unix_peer(other) != sk) {
	 *  -		sock_poll_wait(file, &unix_sk(other)->peer_wait, wait);
	 *  -		if (kgr_unix_recvq_full(other))
	 *  -			writable = 0;
	 *  -	}
	 *  -	sock_put(other);
	 *  +next 10 rows
	 */
	if (writable) {
		unix_state_lock(sk);

		other = unix_peer(sk);
		if (other && unix_peer(other) != sk &&
		    kgr_unix_recvq_full(other) &&
		    kgr_unix_dgram_peer_wake_me(sk, other))
			writable = 0;

		unix_state_unlock(sk);
	}

	if (writable)
		mask |= POLLOUT | POLLWRNORM | POLLWRBAND;
	else
		set_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);

	return mask;
}

static int kgr_patch_bsc955837_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	/* proto_ops/kgr_unix_dgram_ops init */
	kgr_unix_dgram_ops.release = kgr_unix_release; /* patched function */
	kgr_unix_dgram_ops.bind = kgr_unix_bind;
	kgr_unix_dgram_ops.connect = kgr_unix_dgram_connect, /* patched function */
	kgr_unix_dgram_ops.socketpair = kgr_unix_socketpair;
	kgr_unix_dgram_ops.getname = kgr_unix_getname;
	kgr_unix_dgram_ops.poll = kgr_unix_dgram_poll; /* patched function */
	kgr_unix_dgram_ops.ioctl = kgr_unix_ioctl;
	kgr_unix_dgram_ops.shutdown = kgr_unix_shutdown;
	kgr_unix_dgram_ops.sendmsg = kgr_unix_dgram_sendmsg; /* patched function */
	kgr_unix_dgram_ops.recvmsg = kgr_unix_dgram_recvmsg;
	kgr_unix_dgram_ops.set_peek_off = kgr_unix_set_peek_off;

	pr_warn("kgraft-patch: only newly created AF_UNIX DGRAM sockets are invulnerable to CVE-2013-7446\n");

	return 0;
}

int kgr_patch_bsc955837_init(void)
{
	return kgr_patch_bsc955837_kallsyms();
}
